//==============================================================================
// HSIO.c - Hardware Serial port(s)
//==============================================================================

#include "BapLib.h"
#ifdef __cplusplus
extern "C" {
#endif

#include "hsio.h"


//------------------------------------------------------------------------------
// Defines
//------------------------------------------------------------------------------
#define _BA_PIN_MODE_IO_				0
#define _BA_PIN_MODE_FUNCTION_			1

#define _BA_PIN_MODE_INPUT_				0
#define _BA_PIN_MODE_OUTPUT_			1

#define _BA_SCI3_MODE_ASYNC_			0
#define _BA_SCI3_MODE_SYNC_				1

#define _BA_SCI3_BITS_8_				0
#define _BA_SCI3_BITS_7_				1

#define _BA_SCI3_PARITY_DISABLE_		0
#define _BA_SCI3_PARITY_ENABLE_			1

#define _BA_SCI3_PARITY_EVEN_			0
#define _BA_SCI3_PARITY_ODD_			1

#define _BA_SCI3_STOPBITS_1_			0
#define _BA_SCI3_STOPBITS_2_			1

#define _BA_SCI3_MULPROC_DISABLE_		0
#define _BA_SCI3_MULPROC_ENABLE_		1

#define _BA_SCI3_CLOCK_DIV_0_			0
#define _BA_SCI3_CLOCK_DIV_4_			1
#define _BA_SCI3_CLOCK_DIV_16_			2
#define _BA_SCI3_CLOCK_DIV_64_			3

#define _BA_SCI3_BAUDGEN_INTERNAL_		0

void hsio_setbaud (u32 _lbaudrate);
#ifdef _BAP40_
void hsio_setbaud2 (u32 _lbaudrate);
#endif
//------------------------------------------------------------------------------
// Type definitions
//------------------------------------------------------------------------------
#define	HSERQ_SIZE		64	// Assumed to be a power of 2!
typedef struct _hserq
{	
	u8		iStart;		// start pointer
	u8		iEnd;		// end
	u8		abBuf[HSERQ_SIZE];			// our actual buffer
} HSERQ, *PHSERQ;

//------------------------------------------------------------------------------
// Globals
//------------------------------------------------------------------------------

HSERQ	g_HSERQHSO;	// output queue for hardware serial port 1
HSERQ	g_HSERQHSI;	// input queue

//------------------------------------------------------------------------------
// Forward references
//------------------------------------------------------------------------------

//=============================================================================
// Simple Queue functions for adding characters to be processed by Hserin/Hserout
//=============================================================================
inline void ClearQueue(PHSERQ phq) __attribute__((always_inline));
void ClearQueue(PHSERQ phq)
{
	// Initialize the queue to be empty.
	phq->iStart = 0;
	phq->iEnd = 0;
}


// Add a byte to the queue. - was bool but no one checking return value
inline void AddByteToQueue(PHSERQ phq, u8 b, bool fWait) __attribute__((always_inline));

/*bool*/ void  AddByteToQueue(PHSERQ phq, u8 b, bool fWait)
{
	register u8	iOutNext; 
	
	// Calculate the address to save away the byte in.
	iOutNext = (phq->iEnd + 1) & (HSERQ_SIZE-1);		// Again assume power of 2

	while (iOutNext == phq->iStart) {
		// while the queue is full
		if (!fWait)	
			return /*false */;
	}
	phq->abBuf[phq->iEnd] = b;
	phq->iEnd = iOutNext;
	
	return /*true*/;
}

inline word RetrieveByteFromQueue(PHSERQ phq) __attribute__((always_inline));
word RetrieveByteFromQueue(PHSERQ phq)
{
	word	wRet;
	if (phq->iStart == phq->iEnd)
		return 0xffff;
		
	wRet = phq->abBuf[phq->iStart];

	// Increment our index to where the next byte should be from, handle wrap around	
	phq->iStart = (phq->iStart + 1) & (HSERQ_SIZE-1);
	return wRet;
}




//------------------------------------------------------------------------------
// enablehserial - setup the first serial port
//------------------------------------------------------------------------------
void enablehserial (void)
{
	// First initialize the queues used for input and output.
	ClearQueue(&g_HSERQHSO);
	ClearQueue(&g_HSERQHSI);
	
	// set TXD as TXD
	IO.PMR1.BIT.TXD = _BA_PIN_MODE_FUNCTION_;
	// set RXD as RXD
	SCI3.SCR3.BIT.RE = _BA_PIN_MODE_FUNCTION_;
	// set SCK3 as IO pin
	SCI3.SCR3.BIT.CKE = _BA_PIN_MODE_IO_;
	SCI3.SMR.BIT.COM = _BA_PIN_MODE_IO_;
	// set SCK3 as ASYNC pin
	SCI3.SMR.BIT.COM = _BA_SCI3_MODE_ASYNC_;
	SCI3.SMR.BIT.MP = _BA_SCI3_MULPROC_DISABLE_;
	SCI3.SCR3.BIT.RIE = _FALSE_;		// enable interrupts when we receive a character or error
	SCI3.SCR3.BIT.TE = _FALSE_;
	SCI3.SCR3.BIT.RE = _FALSE_;
	SCI3.SCR3.BIT.MPIE = _FALSE_;
	SCI3.SCR3.BIT.TEIE = _FALSE_;	// For now don't need to enable interrupt when transmit complete...
	SCI3.SCR3.BIT.CKE = _BA_SCI3_BAUDGEN_INTERNAL_;
	// enable receiver
	SCI3.SCR3.BIT.RE = _TRUE_;
	SCI3.SCR3.BIT.TE = _TRUE_;
}

//-----------------------------------------------------------------------------
// ENABLEHSERIAL2 - setup the 2nd serial port
//-----------------------------------------------------------------------------
#ifdef _BAP40_
void enablehserial2 (void)
{
	// set TXD_2 as TXD
	IO.PMR1.BIT.TXD2 = _BA_PIN_MODE_FUNCTION_;
	// set RXD_2 as RXD
	SCI3_2.SCR3.BIT.RE = _BA_PIN_MODE_FUNCTION_;
	// set SCK3 as IO pin
	SCI3_2.SCR3.BIT.CKE = _BA_PIN_MODE_IO_;
	SCI3_2.SMR.BIT.COM = _BA_PIN_MODE_IO_;
	// set SCK3 as ASYNC pin
	SCI3_2.SMR.BIT.COM = _BA_SCI3_MODE_ASYNC_;
	SCI3_2.SMR.BIT.MP = _BA_SCI3_MULPROC_DISABLE_;
	SCI3_2.SCR3.BIT.TIE = _FALSE_;
	SCI3_2.SCR3.BIT.RIE = _FALSE_;
	SCI3_2.SCR3.BIT.TE = _FALSE_;
	SCI3_2.SCR3.BIT.RE = _FALSE_;
	SCI3_2.SCR3.BIT.MPIE = _FALSE_;
	SCI3_2.SCR3.BIT.TEIE = _FALSE_;
	SCI3_2.SCR3.BIT.CKE = _BA_SCI3_BAUDGEN_INTERNAL_;
	// enable transmitter
	SCI3_2.SCR3.BIT.TE = _TRUE_;
	// enable receiver
	SCI3_2.SCR3.BIT.RE = _TRUE_;
}
#endif

//-----------------------------------------------------------------------------
// HSETHSERIAL - Initialize the 1st hardware serial port
//-----------------------------------------------------------------------------
void sethserial (u32 _lbaudrate, u8 _bdatabits, u8 _bparity, u8 _bstopbits)
{
	// set port mode
	SCI3.SMR.BIT.CHR = _BA_SCI3_BITS_8_;
	SCI3.SMR.BIT.CHR = _BA_SCI3_BITS_8_;
	SCI3.SMR.BIT.PE = _BA_SCI3_PARITY_DISABLE_;
	SCI3.SMR.BIT.STOP = _BA_SCI3_STOPBITS_1_;
	hsio_setbaud (_lbaudrate);

	// Ok we are now ready to receive something...
	//SCI3.SCR3.BIT.TIE = _TRUE_;		// enable interrupt for transmit...
	//SCI3.SCR3.BIT.TEIE = _TRUE_;	// enable interrupt for transmit end
	SCI3.SCR3.BIT.RIE = _TRUE_;		// enable interrupts when we receive a character or error
	enable();						// make sure interupts are enabled...
}

//-----------------------------------------------------------------------------
// HSETHSERIAL2 - Initialize the 2nd hardware serial port
//-----------------------------------------------------------------------------
#ifdef _BAP40_
void sethserial2 (u32 _lbaudrate, u8 _bdatabits, u8 _bparity, u8 _bstopbits)
{
	// set port mode
	SCI3_2.SMR.BIT.CHR = _BA_SCI3_BITS_8_;
	SCI3_2.SMR.BIT.CHR = _BA_SCI3_BITS_8_;
	SCI3_2.SMR.BIT.PE = _BA_SCI3_PARITY_DISABLE_;
	SCI3_2.SMR.BIT.STOP = _BA_SCI3_STOPBITS_1_;
	hsio_setbaud2 (_lbaudrate);
}
#endif


//-----------------------------------------------------------------------------
// HSEROUT - output a string
//-----------------------------------------------------------------------------
void hserout (u8 *_psdata, u16 _wcount)
{
	u16 wbyteloop;
	for (wbyteloop=0; wbyteloop<_wcount; wbyteloop++) {
		u8	b = *_psdata++;

		// See if the transmitter is active.
		if (SCI3.SSR.BIT.TDRE) {			// If the transmitter is not active
			// enable transmitter
			SCI3.TDR = b;					// and put the first byte out on the output register...
			SCI3.SCR3.BIT.TIE = _TRUE_;		// enable interrupts when ready for next byte to transmit
		} else {
			low (P4);
			// It was active when we started...
			AddByteToQueue(&g_HSERQHSO, b, true);

			if (SCI3.SSR.BIT.TDRE) {				// Says I can stuff a byte out...
					word w;
				if ((w=RetrieveByteFromQueue(&g_HSERQHSO)) != 0xffff) {		
					SCI3.TDR = w & 0xff;			// and put the first byte out on the output register...
					SCI3.SCR3.BIT.TIE = _TRUE_;		// enable interrupts when ready for next byte to transmit
				}
			}
		}
	}
}


//-----------------------------------------------------------------------------
// HSEROUT2 - output a string
//-----------------------------------------------------------------------------
#ifdef _BAP40_
void hserout2 (u8 *_psdata, u16 _wcount)
{
	u16 wbyteloop;

	for (wbyteloop=0; wbyteloop<_wcount; wbyteloop++) {
		while (SCI3_2.SSR.BIT.TDRE != 1);
		SCI3_2.TDR = *_psdata++;
	}
	while (SCI3_2.SSR.BIT.TDRE != 1);
}
#endif

//-----------------------------------------------------------------------------
// HSERIN - Retrieve bytes of data from the hardware serial port.  Will return
//			the number of bytes actually returned.
//-----------------------------------------------------------------------------
u16 hserin (u32 _ltimeout, u8 *_psdata, u16 _rcount, u16 _cterminator)
{
	u16 wbyteloop;
	u8 *_psdataIN = _psdata;
	u32 lTOCnt;
	word	w;

	// rough calculation is the loop while the queue is empty takes about
	// 104 clocks or on Bap28 about 6.5us ... Could speed this up...
	if (_ltimeout != (u32)-1)
		_ltimeout = (_ltimeout * 2)/13;		// Need to update for bap 40...

	for (wbyteloop=0; wbyteloop<_rcount; wbyteloop++) {
		lTOCnt = _ltimeout;
		while ((w = RetrieveByteFromQueue(&g_HSERQHSI)) == 0xffff) 	{
			// Check for timeout...  First pass will be real primitive
			// maybe should integrate the TimerA code to be used here...
			if ((lTOCnt != (u32)-1) && (--lTOCnt == 0))
				goto Timeout;
		}

		// check for terminator.  - changed to word so we don't
		// have characters we can not check for...
		*_psdata++ = w & 0xff;		// save away the character and increment our output pointer.
		if ((_cterminator>>8 == 0) && (w == _cterminator))
			break;
	}
	
	// return how many characters that 
Timeout:	
	return (u16)(_psdata - _psdataIN);
}

//-----------------------------------------------------------------------------
// HSERIN2
//-----------------------------------------------------------------------------
#ifdef _BAP40_
void hserin2 (u32 _ltimeout, u8 *_psdata, u16 _rcount, u8 _cterminator)
{
	u16 wbyteloop;

	for (wbyteloop=0; wbyteloop<_rcount; wbyteloop++) 	{
		if ((SCI3_2.SSR.BIT.OER == 1) || (SCI3_2.SSR.BIT.PER == 1) || (SCI3_2.SSR.BIT.FER == 1))
			break;
		while (SCI3_2.SSR.BIT.RDRF == 0);
		*_psdata = SCI3_2.RDR;
		if ((_cterminator != 0) && (*_psdata++ == _cterminator))
			break;
	}
}
#endif

// set the baud rate
void hsio_setbaud (u32 _lbaudrate)
{
	// Lets calculate it, not hard code it...
#ifdef _BAP28_	
	// BUGBUG:: should define clockspeed somewhere then we can simplify this...
	word n = (ulong)16000000/_lbaudrate;		// See which divisor we should use
#endif	
#ifdef _BAP40_	
	word n = (ulong)20000000/_lbaudrate;
#endif	
	// see if we can use DIV_0
	if (n < (255*32)) {
		SCI3.SMR.BIT.CKS = _BA_SCI3_CLOCK_DIV_0_;
		SCI3.BRR = (n+16)/32 - 1;
	} else {
		// I think we won't need to go to any higher
		SCI3.SMR.BIT.CKS = _BA_SCI3_CLOCK_DIV_4_;
		SCI3.BRR = (n+64)/128 - 1;
	}
}

#ifdef _BAP40_
// set the baud rate
void hsio_setbaud2 (u32 _lbaudrate)
{
	SCI3_2.SMR.BIT.CKS = _BA_SCI3_CLOCK_DIV_0_;
	switch (_lbaudrate)
	{
		case 1200:
			SCI3_2.SMR.BIT.CKS = _BA_SCI3_CLOCK_DIV_4_;
			SCI3_2.BRR = 129;
			break;

		case 2400:
			SCI3_2.SMR.BIT.CKS = _BA_SCI3_CLOCK_DIV_4_;
			SCI3_2.BRR = 64;
			break;

		case 4800:
			SCI3_2.BRR = 120;
			break;

		case 9600:
			SCI3_2.BRR = 64;
			break;

		case 19200:
			SCI3_2.BRR = 32;
			break;

		case 31250:
			SCI3_2.BRR = 19;
			break;

		case 38400:
			SCI3_2.BRR = 15;
			break;

		case 57600:
			SCI3_2.BRR = 10;
			break;

		case 115200:
			SCI3_2.BRR = 5;
			break;
	}
}
#endif


//-----------------------------------------------------------------------------
// HSERSTAT
//-----------------------------------------------------------------------------
bool hserstat(HSERSTAT_OPTIONS ho)  // ? - would it be better to simply have a function for each???
{
	switch (ho)
	{
		case HSERSTAT_CLEAR_INPUT_BUFFER:
			ClearQueue(&g_HSERQHSI);
			return true;

		case HSERSTAT_CLEAR_BOTH_BUFFERS:
			ClearQueue(&g_HSERQHSI);
			// fall through			
		case HSERSTAT_CLEAR_OUTPUT_BUFFER:
			ClearQueue(&g_HSERQHSO);
			return true;
			
		case HSERSTAT_INPUT_AVAILABLE:
			return (g_HSERQHSI.iStart != g_HSERQHSI.iEnd);
			
		case HSERSTAT_OUTPUT_EMPTY:
			return (g_HSERQHSO.iStart == g_HSERQHSO.iEnd);
	}

	return false;	// unknown command
}


//=============================================================================
// Interrupt handler for SCI3 (UART)
//=============================================================================
extern "C" void SCI3INT() __attribute__ ((interrupt_handler));
extern "C" void SCI3INT()
{
	word	w;
	// OK we need to find out why we were interrupted...
	toggle (P5);
	if (SCI3.SSR.BIT.TDRE) 	{
		// Transmit Data register empty
		if ((w=RetrieveByteFromQueue(&g_HSERQHSO)) != 0xffff)
			SCI3.TDR = w & 0xff;
		else
			SCI3.SCR3.BIT.TIE = _FALSE_;		// Disable this interrupt when ready for next byte to transmit
			
	}

	if (SCI3.SSR.BIT.OER) {
		// Data overrun error
		SCI3.SSR.BIT.OER = 0; 		// should maybe do something with error conditions?
	}

	if (SCI3.SSR.BIT.RDRF) {
		// Receive data Register Full
		// Should maybe do something if our queue was full and we errored...
		w = SCI3.RDR;
		AddByteToQueue(&g_HSERQHSI,  w &0xff, false);
	}

	if (SCI3.SSR.BIT.FER) {
		// data framing error
		SCI3.SSR.BIT.FER = 0;
	}
	
	if (SCI3.SSR.BIT.PER) {
		// data parity error
		SCI3.SSR.BIT.PER = 0;
	}
	toggle (P5);
}


#ifdef __cplusplus
}
#endif
